@@ -1,7 +1,9 @@
 import os
 from openwakeword.model import Model
+from openwakeword.vad import VAD
+from openwakeword.custom_verifier_model import train_custom_verifier
 
-__all__ = ['Model', ]
+__all__ = ['Model', 'VAD', 'train_custom_verifier']
 
 models = {
     "alexa": {
@@ -10,6 +12,9 @@ models = {
     "hey_mycroft": {
         "model_path": os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/models/hey_mycroft_v0.1.onnx")
     },
+    "hey_jarvis": {
+        "model_path": os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/models/hey_jarvis_v0.1.onnx")
+    },
     "timer": {
         "model_path": os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/models/timer_v0.1.onnx")
     },
