@@ -19,7 +19,7 @@ from functools import partial
 from pathlib import Path
 import random
 from tqdm import tqdm
-from typing import List
+from typing import List, Tuple
 import numpy as np
 import torch
 from numpy.lib.format import open_memmap
@@ -27,6 +27,7 @@ from speechbrain.dataio.dataio import read_audio
 from speechbrain.processing.signal_processing import reverberate
 import torchaudio
 import mutagen
+import acoustics
 
 
 # Load audio clips and structure into clips of the same length
@@ -224,6 +225,40 @@ def estimate_clip_duration(audio_files: list, sizes: list):
     return durations
 
 
+def estimate_mp3_duration(fpath):
+    """Estimates the duration of an MP3 file from metadata and file-size.
+    Is only accurate for 16000 khz sample rate audio with a relatively
+    constant bit-rate.
+
+    Args:
+        fpath (str): The input path to the MP3 file
+
+    Returns:
+        float: The duration of the MP3 file in seconds
+    """
+
+    conversion_factors = {
+        "16_khz_single_channel": 0.000333318208471784,
+        "16_khz_stereo": 0.000333318208471784/2
+    }
+
+    duration_seconds = 0
+    try:
+        md = torchaudio.info(fpath)
+    except RuntimeError:
+        return duration_seconds
+
+    nbytes = os.path.getsize(fpath)
+    if md.num_channels == 1:
+        if md.sample_rate == 16000:
+            duration_seconds = nbytes*conversion_factors["16_khz_single_channel"]
+    elif md.num_channels == 2:
+        if md.sample_rate == 16000:
+            duration_seconds = nbytes*conversion_factors["16_khz_stereo"]
+
+    return duration_seconds
+
+
 def get_clip_duration(clip):
     """Gets the duration of an audio clip in seconds from file header information"""
     try:
@@ -259,8 +294,16 @@ def mix_clips_batch(
         snr_low: float = 0,
         snr_high: float = 0,
         start_index: List[int] = [],
+        foreground_durations: List[float] = [],
+        foreground_truncate_strategy: str = "random",
         rirs: List[str] = [],
+        rir_probability: int = 1,
+        volume_augmentation: bool = True,
+        generated_noise_augmentation: float = 0.0,
         shuffle: bool = True,
+        return_sequence_labels: bool = False,
+        return_background_clips: bool = False,
+        return_background_clips_delay: Tuple[int, int] = (0, 0),
         seed: int = 0
         ):
     """
@@ -281,24 +324,53 @@ def mix_clips_batch(
         snr_low (float): The low SNR level of the mixing in db
         snr_high (float): The high snr level of the mixing in db
         start_index (List[int]): The starting position (in samples) for the foreground clip to start in
-                                 the background clip.
+                                 the background clip. If the foreground clip is longer than `combined_size`
+                                 when starting at this point, the foreground clip will be truncated
+                                 according to the `foreground_truncate_strategy` argument.
+        foreground_durations (List[float]): The desired duration of each foreground clip (in seconds)
+        foreground_truncate_strategy (str): The method used to truncate the foreground clip, if needed based on the
+                                            `start_index`, `foreground_durations`, and `combined_size` arguments.
+                                            See the options in the `truncate_clip` method.
         rirs (List[str]): A list of paths to room impulse response functions (RIR) to convolve with the
-                          clips to simulate different recording environments. Applies a single random from the
+                          clips to simulate different recording environments. Applies a single random selection from the
                           list RIR file to the entire batch. If empty (the default), nothing is done.
+        rir_probability (float): The probability (between 0 and 1) that the batch will be convolved with a RIR file.
+        volume_augmentation (bool): Whether to randomly apply volume augmentation to the clips in the batch.
+                                    This simply scales the data of each clip such that the maximum value is is between
+                                    0.02 and 1.0 (the floor shouldn't be zero as beyond a certain point the audio data
+                                    is no longer valid).
+        generated_noise_augmentation: The probability of further mixing the mixed clip with generated random noise.
+                                      Will be either "white", "brown", "blue", "pink", or "violet" noise, mixed at a
+                                      random SNR between `snr_low` and `snr_high`.
+        return_background_clips (bool): Whether to return the segment of the background clip that was mixed with each
+                                        foreground clip in the batch.
+        return_background_clips_delay (Tuple(int)): The lower and upper bound of a random delay (in samples)
+                                           to apply to the segment of each returned backgroud clip mixed
+                                           with each foreground clip in the batch. This is primarily intended to
+                                           simulate the drift between input and output channels
+                                           in audio devices, which means that the mixed audio is never
+                                           exactly aligned with the two source clips.
         shuffle (bool): Whether to shuffle the foreground clips before mixing (default: True)
+        return_sequence_labels (bool): Whether to return sequence labels (i.e., frame-level labels) for each clip
+                                       based on the start/end positions of the foreground clip.
         seed (int): A random seed
 
     Returns:
-        generator: Returns a generator that yields batches of mixed foreground/background audio
+        generator: Returns a generator that yields batches of mixed foreground/background audio, labels, and the
+                   background segments used for each audio clip (or None is the
+                   `return_backgroun_clips` argument is False)
     """
     # Set random seed, if needed
     if seed:
         np.random.seed(seed)
         random.seed(seed)
 
-    # Set start indices, if needed
+    # Check and Set start indices, if needed
     if not start_index:
         start_index = [0]*batch_size
+    else:
+        if min(start_index) < 0:
+            raise ValueError("Error! At least one value of the `start_index` argument is <0. Check your inputs.")
 
     # Make dummy labels
     if not labels:
@@ -309,49 +381,77 @@ def mix_clips_batch(
         foreground_clips = np.array(foreground_clips)[p].tolist()
         start_index = np.array(start_index)[p].tolist()
         labels = np.array(labels)[p].tolist()
+        if foreground_durations:
+            foreground_durations = np.array(foreground_durations)[p].tolist()
 
     for i in range(0, len(foreground_clips), batch_size):
-        # Load foreground clips/start indices and truncate (if needed)
-        foreground_clips_batch = [read_audio(i)[0:combined_size] for i in foreground_clips[i:i+batch_size]]
+        # Load foreground clips/start indices and truncate as needed
+        sr = 16000
         start_index_batch = start_index[i:i+batch_size]
+        foreground_clips_batch = [read_audio(j) for j in foreground_clips[i:i+batch_size]]
+        foreground_clips_batch = [j[0] if len(j.shape) > 1 else j for j in foreground_clips_batch]
+        if foreground_durations:
+            foreground_clips_batch = [truncate_clip(j, int(k*sr), foreground_truncate_strategy)
+                                      for j, k in zip(foreground_clips_batch, foreground_durations[i:i+batch_size])]
         labels_batch = np.array(labels[i:i+batch_size])
 
         # Load background clips and pad/truncate as needed
-        background_clips_batch = [read_audio(i) for i in random.sample(background_clips, batch_size)]
+        background_clips_batch = [read_audio(j) for j in random.sample(background_clips, batch_size)]
+        background_clips_batch = [j[0] if len(j.shape) > 1 else j for j in background_clips_batch]
+        background_clips_batch_delayed = []
+        delay = np.random.randint(return_background_clips_delay[0], return_background_clips_delay[1] + 1)
         for ndx, background_clip in enumerate(background_clips_batch):
-            if background_clip.shape[0] < combined_size:
-                background_clips_batch[ndx] = background_clip.repeat(
-                    np.ceil(combined_size/background_clip.shape[0]).astype(np.int32)
-                )[0:combined_size]
-            elif background_clip.shape[0] > combined_size:
-                r = np.random.randint(0, max(1, background_clip.shape[0] - combined_size))
+            if background_clip.shape[0] < (combined_size + delay):
+                repeated = background_clip.repeat(
+                    np.ceil((combined_size + delay)/background_clip.shape[0]).astype(np.int32)
+                )
+                background_clips_batch[ndx] = repeated[0:combined_size]
+                background_clips_batch_delayed.append(repeated[0+delay:combined_size + delay].clone())
+            elif background_clip.shape[0] > (combined_size + delay):
+                r = np.random.randint(0, max(1, background_clip.shape[0] - combined_size - delay))
                 background_clips_batch[ndx] = background_clip[r:r + combined_size]
+                background_clips_batch_delayed.append(background_clip[r+delay:r + combined_size + delay].clone())
 
         # Mix clips at snr levels
         snrs_db = np.random.uniform(snr_low, snr_high, batch_size)
         mixed_clips = []
+        sequence_labels = []
         for fg, bg, snr, start in zip(foreground_clips_batch, background_clips_batch,
                                       snrs_db, start_index_batch):
-            fg_rms, bg_rms = fg.norm(p=2), bg.norm(p=2)
-            snr = 10 ** (snr / 20)
-            scale = snr * bg_rms / fg_rms
-            bg[start:start + fg.shape[0]] = bg[start:start + fg.shape[0]] + scale*fg
-            mixed_clips.append(bg / 2)
+            if bg.shape[0] != combined_size:
+                raise ValueError(bg.shape)
+            mixed_clip = mix_clip(fg, bg, snr, start)
+            sequence_labels.append(get_frame_labels(combined_size, start, start+fg.shape[0]))
+
+            if np.random.random() < generated_noise_augmentation:
+                noise_color = ["white", "pink", "blue", "brown", "violet"]
+                noise_clip = acoustics.generator.noise(combined_size, color=np.random.choice(noise_color))
+                noise_clip = torch.from_numpy(noise_clip/noise_clip.max())
+                mixed_clip = mix_clip(mixed_clip, noise_clip, np.random.choice(snrs_db), 0)
+
+            mixed_clips.append(mixed_clip)
 
         mixed_clips_batch = torch.vstack(mixed_clips)
+        sequence_labels_batch = torch.from_numpy(np.vstack(sequence_labels))
 
         # Apply reverberation to the batch (from a single RIR file)
         if rirs:
-            rir_waveform, sr = torchaudio.load(random.choice(rirs))
-            if rir_waveform.shape[0] > 1:
-                rir_waveform = rir_waveform[random.randint(0, rir_waveform.shape[0]-1), :]
-            mixed_clips_batch = reverberate(mixed_clips_batch, rir_waveform, rescale_amp="avg")
+            if np.random.random() <= rir_probability:
+                rir_waveform, sr = torchaudio.load(random.choice(rirs))
+                if rir_waveform.shape[0] > 1:
+                    rir_waveform = rir_waveform[random.randint(0, rir_waveform.shape[0]-1), :]
+                mixed_clips_batch = reverberate(mixed_clips_batch, rir_waveform, rescale_amp="avg")
 
-        # Normalize clips only if max value is outside of [-1, 1]
-        abs_max, _ = torch.max(
-            torch.abs(mixed_clips_batch), dim=1, keepdim=True
-        )
-        mixed_clips_batch = mixed_clips_batch / abs_max.clamp(min=1.0)
+        # Apply volume augmentation
+        if volume_augmentation:
+            volume_levels = np.random.uniform(0.02, 1.0, mixed_clips_batch.shape[0])
+            mixed_clips_batch = (volume_levels/mixed_clips_batch.max(axis=1)[0])[..., None]*mixed_clips_batch
+        else:
+            # Normalize clips only if max value is outside of [-1, 1]
+            abs_max, _ = torch.max(
+                torch.abs(mixed_clips_batch), dim=1, keepdim=True
+            )
+            mixed_clips_batch = mixed_clips_batch / abs_max.clamp(min=1.0)
 
         # Convert to 16-bit PCM audio
         mixed_clips_batch = (mixed_clips_batch.numpy()*32767).astype(np.int16)
@@ -360,8 +460,65 @@ def mix_clips_batch(
         error_index = np.where(mixed_clips_batch.max(axis=1) != 0)[0]
         mixed_clips_batch = mixed_clips_batch[error_index]
         labels_batch = labels_batch[error_index]
+        sequence_labels_batch = sequence_labels_batch[error_index]
 
-        yield mixed_clips_batch, labels_batch
+        if not return_background_clips:
+            yield mixed_clips_batch, labels_batch if not return_sequence_labels else sequence_labels_batch, None
+        else:
+            background_clips_batch_delayed = (torch.vstack(background_clips_batch_delayed).numpy()
+                                              * 32767).astype(np.int16)[error_index]
+            yield (mixed_clips_batch,
+                   labels_batch if not return_sequence_labels else sequence_labels_batch,
+                   background_clips_batch_delayed)
+
+
+def get_frame_labels(combined_size, start, end, buffer=1):
+    sequence_label = np.zeros(np.ceil((combined_size-12400)/1280).astype(int))
+    frame_positions = np.arange(12400, combined_size, 1280)
+    start_frame = np.argmin(abs(frame_positions - start))
+    end_frame = np.argmin(abs(frame_positions - end))
+    sequence_label[start_frame:start_frame+2] = 1
+    sequence_label[end_frame-1:end_frame+1] = 1
+    return sequence_label
+
+
+def mix_clip(fg, bg, snr, start):
+    fg_rms, bg_rms = fg.norm(p=2), bg.norm(p=2)
+    snr = 10 ** (snr / 20)
+    scale = snr * bg_rms / fg_rms
+    bg[start:start + fg.shape[0]] = bg[start:start + fg.shape[0]] + scale*fg
+    return bg / 2
+
+
+def truncate_clip(x, max_size, method="truncate_start"):
+    """
+    Truncates and audio clip with the specified method
+
+    Args:
+        x (nd.array): An array of audio data
+        max_size (int): The maximum size (in samples)
+        method (str): Can be one of four options:
+            - "truncate_start": Truncate the start of the clip
+            - "truncate_end": Truncate the end of the clip
+            - "truncate_both": Truncate both the start and end of the clip
+            - "random": Randomly select a segment of the right size from the clip
+
+    Returns:
+        nd.array: The truncated audio data
+    """
+    if x.shape[0] > max_size:
+        if method == "truncate_start":
+            x = x[x.shape[0] - max_size:]
+        if method == "truncate_end":
+            x = x[0:max_size]
+        if method == "truncate_both":
+            n = int(np.ceil(x.shape[0] - max_size)/2)
+            x = x[n:-n][0:max_size]
+        if method == "random":
+            rn = np.random.randint(0, x.shape[0] - max_size)
+            x = x[rn:rn + max_size]
+
+    return x
 
 
 # Reverberation data augmentation function
@@ -402,7 +559,8 @@ class mmap_batch_generator:
     """
     def __init__(self,
                  data_files: dict,
-                 batch_size: int,
+                 label_files: dict = {},
+                 batch_size: int = 128,
                  n_per_class: dict = {},
                  data_transform_funcs: dict = {},
                  label_transform_funcs: dict = {}
@@ -413,6 +571,9 @@ class mmap_batch_generator:
         Args:
             data_files (dict): A dictionary of labels (as keys) and on-disk numpy array paths (as values).
                                Keys should be integer strings representing class labels.
+            label_files (dict): A dictionary where the keys are the class labels and the values are the per-example
+                                labels. The values must be the same shape as the correponding numpy data arrays
+                                from the `data_files` argument.
             batch_size (int): The number of samples per batch
             n_per_class (dict): A dictionary with integer string labels (as keys) and number of example per batch
                                (as values). If None (the default), batch sizes for each class will be
@@ -437,12 +598,14 @@ class mmap_batch_generator:
         """
         # inputs
         self.data_files = data_files
+        self.label_files = label_files
         self.n_per_class = n_per_class
         self.data_transform_funcs = data_transform_funcs
         self.label_transform_funcs = label_transform_funcs
 
         # Get array mmaps and store their shapes (but load files < 1 GB total size into memory)
         self.data = {label: np.load(fl, mmap_mode='r') for label, fl in data_files.items()}
+        self.labels = {label: np.load(fl) for label, fl in label_files.items()}
         self.data_counter = {label: 0 for label in data_files.keys()}
         self.original_shapes = {label: self.data[label].shape for label in self.data.keys()}
         self.shapes = {label: self.data[label].shape for label in self.data.keys()}
@@ -454,22 +617,22 @@ class mmap_batch_generator:
         #     self.shapes[lbl] = (new_shape[0]*self.original_shapes[lbl][0], new_shape[1], new_shape[2])
 
         # Calculate batch sizes, if the user didn't specify them
+        scale_factor = 1
         if not self.n_per_class:
             self.n_per_class = {}
             for lbl, shape in self.shapes.items():
                 dummy_data = np.random.random((10, self.shapes[lbl][1], self.shapes[lbl][2]))
                 if self.data_transform_funcs.get(lbl, None):
                     scale_factor = self.data_transform_funcs.get(lbl, None)(dummy_data).shape[0]/10
-                else:
-                    scale_factor = 1
+
                 ratio = self.shapes[lbl][0]/sum([i[0] for i in self.shapes.values()])
                 self.n_per_class[lbl] = max(1, int(int(batch_size*ratio)/scale_factor))
 
-        # Get estimated batches per epoch, including the effect of any user-provided transforms
-        batch_size = sum([val*scale_factor for val in self.n_per_class.values()])
-        batches_per_epoch = sum([i[0] for i in self.shapes.values()])//batch_size
-        self.batch_per_epoch = batches_per_epoch
-        print("Batches/steps per epoch:", batches_per_epoch)
+            # Get estimated batches per epoch, including the effect of any user-provided transforms
+            batch_size = sum([val*scale_factor for val in self.n_per_class.values()])
+            batches_per_epoch = sum([i[0] for i in self.shapes.values()])//batch_size
+            self.batch_per_epoch = batches_per_epoch
+            print("Batches/steps per epoch:", batches_per_epoch)
 
     def __iter__(self):
         return self
@@ -493,7 +656,10 @@ class mmap_batch_generator:
                     x = self.data_transform_funcs[label](x)
 
                 # Make labels for data (following whatever the current shape of `x` is)
-                y_batch = [label]*x.shape[0]
+                if self.label_files.get(label, None):
+                    y_batch = self.labels[label][self.data_counter[label]:self.data_counter[label]+n]
+                else:
+                    y_batch = [label]*x.shape[0]
 
                 # Transform labels
                 if self.label_transform_funcs and self.label_transform_funcs.get(label):
